# copyright 2004-2014 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of yams.
#
# yams is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# yams is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with yams. If not, see <http://www.gnu.org/licenses/>.
"""ER schema loader.

Use either a sql derivated language for entities and relation definitions
files or a direct python definition file.
"""
from __future__ import print_function

__docformat__ = "restructuredtext en"

import sys
import types
from os import listdir
from os.path import (dirname, exists, join, splitext, basename, abspath,
                     realpath)
from warnings import warn

from logilab.common import tempattr
from logilab.common.modutils import modpath_from_file, cleanup_sys_modules

from yams import UnknownType, BadSchemaDefinition, BASE_TYPES
from yams import constraints, schema as schemamod
from yams import buildobjs


CONSTRAINTS = {}
# add constraint classes to the context
for objname in dir(constraints):
    if objname[0] == '_':
        continue
    obj = getattr(constraints, objname)
    try:
        if issubclass(obj, constraints.BaseConstraint) and (
            not obj is constraints.BaseConstraint):
            CONSTRAINTS[objname] = obj
    except TypeError:
        continue


def fill_schema(schema, erdefs, register_base_types=True,
                remove_unused_rtypes=False, post_build_callbacks=[]):
    if register_base_types:
        buildobjs.register_base_types(schema)
    # relation definitions may appear multiple times
    erdefs_vals = set(erdefs.values())
    # register relation types and non final entity types
    for definition in erdefs_vals:
        if isinstance(definition, type):
            definition = definition()
        if isinstance(definition, buildobjs.RelationType):
            schema.add_relation_type(definition)
        elif isinstance(definition, buildobjs.EntityType):
            schema.add_entity_type(definition)
    # register relation definitions
    for definition in erdefs_vals:
        if isinstance(definition, type):
            definition = definition()
        definition.expand_relation_definitions(erdefs, schema)
    # call 'post_build_callback' functions found in schema modules
    for cb in post_build_callbacks:
        cb(schema)
    # finalize schema
    schema.finalize()
    # check permissions are valid on entities and relations
    for erschema in schema.entities() + schema.relations():
        erschema.check_permission_definitions()
    # check unique together consistency
    for eschema in schema.entities():
        eschema.check_unique_together()
    # optionaly remove relation types without definitions
    if remove_unused_rtypes:
        for rschema in schema.relations():
            if not rschema.rdefs:
                schema.del_relation_type(rschema)
    return schema


class SchemaLoader(object):
    """the schema loader is responsible to build a schema object from a
    set of files
    """
    schemacls = schemamod.Schema
    extrapath = None
    context = dict([(attr, getattr(buildobjs, attr))
                    for attr in buildobjs.__all__])
    context.update(CONSTRAINTS)

    def load(self, directories, name=None,
             register_base_types=True, construction_mode='strict',
             remove_unused_rtypes=True):
        """return a schema from the schema definition read from <directory>
        """
        self.defined = {}
        self.loaded_files = []
        self.post_build_callbacks = []
        sys.modules[__name__].context = self
        # ensure we don't have an iterator
        directories = tuple(directories)
        try:
            self._load_definition_files(directories)
            schema = self.schemacls(name or 'NoName', construction_mode=construction_mode)
            try:
                fill_schema(schema, self.defined, register_base_types,
                            remove_unused_rtypes=remove_unused_rtypes,
                            post_build_callbacks=self.post_build_callbacks)
            except Exception as ex:
                if not hasattr(ex, 'schema_files'):
                    ex.schema_files = self.loaded_files
                raise
        finally:
            # cleanup sys.modules from schema modules
            # ensure we're only cleaning schema [sub]modules
            directories = [(not directory.endswith(self.main_schema_directory)
                            and join(directory, self.main_schema_directory)
                            or directory)
                           for directory in directories]
            cleanup_sys_modules(directories)
        schema.loaded_files = self.loaded_files
        return schema

    def _load_definition_files(self, directories):
        for directory in directories:
            package = basename(directory)
            for filepath in self.get_schema_files(directory):
                with tempattr(buildobjs, 'PACKAGE', package):
                    self.handle_file(filepath)

    # has to be overridable sometimes (usually for test purpose)
    main_schema_directory = 'schema'
    def get_schema_files(self, directory):
        """return an ordered list of files defining a schema

        look for a schema.py file and or a schema sub-directory in the given
        directory
        """
        result = []
        if exists(join(directory, self.main_schema_directory + '.py')):
            result = [join(directory, self.main_schema_directory + '.py')]
        if exists(join(directory, self.main_schema_directory)):
            directory = join(directory, self.main_schema_directory)
            for filename in listdir(directory):
                if filename[0] == '_':
                    if filename == '__init__.py':
                        result.insert(0, join(directory, filename))
                    continue
                ext = splitext(filename)[1]
                if ext == '.py':
                    result.append(join(directory, filename))
                else:
                    self.unhandled_file(join(directory, filename))
        return result

    def handle_file(self, filepath):
        """handle a partial schema definition file according to its extension
        """
        assert filepath.endswith('.py'), 'not a python file'
        if filepath not in self.loaded_files:
            modname, module = self.exec_file(filepath)
            objects_to_add = set()
            for name, obj in vars(module).items():
                if (isinstance(obj, type)
                    and issubclass(obj, buildobjs.Definition)
                    and obj.__module__ == modname
                    and not name.startswith('_')):
                    objects_to_add.add(obj)
            for obj in objects_to_add:
                self.add_definition(obj, filepath)
            self.loaded_files.append(filepath)

    def unhandled_file(self, filepath):
        """called when a file without handler associated has been found,
        does nothing by default.
        """
        pass

    def add_definition(self, defobject, filepath=None):
        """file handler callback to add a definition object

        wildcard capability force to load schema in two steps : first register
        all definition objects (here), then create actual schema objects (done in
        `_build_schema`)
        """
        if not issubclass(defobject, buildobjs.Definition):
            raise BadSchemaDefinition(filepath, 'invalid definition object')
        defobject.expand_type_definitions(self.defined)

    def exec_file(self, filepath):
        try:
            modname = '.'.join(modpath_from_file(filepath, self.extrapath))
            doimport = True
        except ImportError:
            warn('module for %s can\'t be found, add necessary __init__.py '
                 'files to make it importable' % filepath, DeprecationWarning)
            modname = splitext(basename(filepath))[0]
            doimport = False
        # XXX can't rely on __import__ until bw compat (eg implicit import) needed
        #if doimport:
        #    module = __import__(modname, fglobals)
        #    for part in modname.split('.')[1:]:
        #        module = getattr(module, part)
        #else:
        if modname in sys.modules:
            module = sys.modules[modname]
            # NOTE: don't test raw equality to avoid .pyc / .py comparisons
            mpath = realpath(abspath(module.__file__))
            fpath = realpath(abspath(filepath))
            assert mpath.startswith(fpath), (
                modname, filepath, module.__file__)
        else:
            fglobals = {} # self.context.copy()
            fglobals['__file__'] = filepath
            fglobals['__name__'] = modname
            package = '.'.join(modname.split('.')[:-1])
            if package and not package in sys.modules:
                __import__(package)
            with open(filepath) as f:
                try:
                    code = compile(f.read(), filepath, 'exec')
                    exec(code, fglobals)
                except:
                    print('exception while reading %s' % filepath, file=sys.stderr)
                    raise
            fglobals['__file__'] = filepath
            module = types.ModuleType(str(modname))
            module.__dict__.update(fglobals)
            sys.modules[modname] = module
            if package:
                setattr(sys.modules[package], modname.split('.')[-1], module)
        if hasattr(module, 'post_build_callback'):
            self.post_build_callbacks.append(module.post_build_callback)
        return (modname, module)

# XXX backward compatibility to prevent changing cw.schema and cw.test.unittest_schema (3.12.+)
PyFileReader = SchemaLoader
PyFileReader.__init__ = lambda *x: None

def fill_schema_from_namespace(schema, items, **kwargs):
    erdefs = {}
    for name, obj in items:
        if (isinstance(obj, type) and issubclass(obj, buildobjs.Definition)
            and obj not in (buildobjs.Definition, buildobjs.RelationDefinition, buildobjs.EntityType)):
            obj.expand_type_definitions(erdefs)
    fill_schema(schema, erdefs, **kwargs)

def build_schema_from_namespace(items):
    schema = schemamod.Schema('noname')
    fill_schema_from_namespace(schema, items)
    return schema

class _Context(object):
    def __init__(self):
        self.defined = {}

context = _Context()



# copyright 2013 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of yams.
#
# yams is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# yams is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with yams. If not, see <http://www.gnu.org/licenses/>.
"""Compare two yams schemas

Textual representation of schema are created and standard diff algorithm are
applied.
"""
from __future__ import print_function

import subprocess
import tempfile
import os

from six import string_types, text_type

from yams.constraints import (SizeConstraint,
                              UniqueConstraint,
                              StaticVocabularyConstraint)
from yams.reader import SchemaLoader


def quoted(obj):
    if isinstance(obj, string_types):
        # NOT repr, because .....
        # strings here turn out as unicode strings (from the repo side)
        # but are (mostly) really str in the schema.py source
        # and we don't want a bogus diff about '...' vs u'...'
        if "'" in obj:
            return '"%s"' % obj
    return "'%s'" % obj

class attr_scope(object):
    indentation = 2
    spacer = ''
    nextline = ','

class class_scope(object):
    indentation = 1
    spacer = ' '
    nextline = ''

def identity(x):
    return x

def format_props(props, scope, ignore=()):
    out = []
    for prop, val in sorted(props.items()):
        if prop in ignore:
            continue
        out.append('\t' * scope.indentation +
                   '%s%s=%s%s' % (prop, scope.spacer, scope.spacer, val))
    return ('%s\n' % scope.nextline).join(out)

def format_expression(expr):
    mainvars = expr.mainvars - set('SUXO')
    return '%s(%s%s)' % (expr.__class__.__name__,
                         quoted(expr.expression),
                         ','.join(mainvars) if mainvars else '')

def format_tuple(iterator):
    out = []
    for thing in iterator:
        out.append(thing)
    return '(%s)' % ','.join(out)

def format_perms(perms, scope, isdefault):
    out = []
    for p in ('read', 'add', 'update', 'delete'):
        rule = perms.get(p)
        if rule:
            out.append('%s%s: %s' %
                       ('\t' + '\t' * scope.indentation,
                        quoted(p),
                        format_tuple(format_expression(r)
                                     if not isinstance(r, string_types) else quoted(r)
                              for r in rule)))
    return ' {%s\n%s\n%s}' % (' # default perms' if isdefault else '',
                              ',\n'.join(out),
                              '\t' * scope.indentation)

def format_constraint(cstr):
    cclass = cstr.__class__.__name__
    if 'RQL' in cclass:
        out = ['%s(%s' % (cclass, quoted(cstr.expression))]
        mainvars = getattr(cstr, 'mainvars', None)
        if mainvars:
            out.append(', mainvars=%s' % quoted(','.join(mainvars)))
        out.append(')')
    elif 'Boundary' in cclass:
        return '%s(%s %s)' % (cclass, quoted(cstr.operator), str(cstr.boundary))
    else:
        print('formatter: unhandled constraint type', cstr.__class__)
        return str(cstr)
    return ''.join(out)

def nullhandler(relation, perms):
    return perms, False

def properties_from(relation, permissionshandler=nullhandler):
    """return a dictionary containing properties of an attribute
    or a relation (if specified with is_relation option)"""
    ret = {}
    for prop in relation.rproperties():
        if prop == 'infered':
            continue

        val = getattr(relation, prop)
        if prop == 'permissions':
            val, default = permissionshandler(relation, val)
            if val:
                ret['__permissions__'] = format_perms(val, attr_scope, default)
            continue
        elif prop == 'cardinality' and relation.final:
            # attributes: re-sugar cardinality to 'required' prop
            prop = 'required'
            val = val[0] == '1'
            if not val:
                continue # don't emit 'required = False' props
        elif prop in ('indexed', 'fulltextindexed', 'uid', 'internationalizable'):
            if not val:
                continue # don't emit 'prop = False' props
        elif prop == 'constraints':
            other = []
            for constraint in val:
                # first, resugar some classic constraints
                if isinstance(constraint, SizeConstraint):
                    ret['maxsize'] = constraint.max
                elif isinstance(constraint, UniqueConstraint):
                    ret['unique'] = True
                elif isinstance(constraint, StaticVocabularyConstraint):
                    if relation.object.type == 'String':
                        transform = text_type
                    else:
                        transform = identity
                    ret['vocabulary'] = [transform(x)
                                         for x in constraint.vocabulary()]
                else:
                    other.append(constraint)
            if other:
                ret['constraints'] = '[%s]' % ','.join(format_constraint(cstr) for cstr in other)
            continue

        if val is None:
            continue

        if prop == 'default':
            if not isinstance(val, bool):
                if not val:
                    continue # default = ''
        elif prop in ('description', 'cardinality', 'composite'):
            if val:
                val = quoted(val)
            else: # let's skip empty strings
                continue

        ret[prop] = val
    return ret

def schema2descr(schema, permissionshandler, ignore=()):
    """convert a yams schema into a text description"""
    output = []
    ignore = set(ignore)
    multirtypes = []
    for entity in sorted(schema.entities()):
        if entity.final:
            continue
        output.append('class %s(EntityType):\n' % entity.type )
        perms, isdefault = permissionshandler(entity, entity.permissions)
        output.append('\t__permissions__ = %s\n\n' %format_perms(perms, class_scope, isdefault))

        attributes = [(attr[0].type,
                       attr[1].type,
                       properties_from(entity.rdef(attr[0].type), permissionshandler))
                      for attr in sorted(entity.attribute_definitions())]
        for attr_name, attr_type, attr_props in attributes:
            if attr_name in ignore:
                continue
            output.append('\t%s = %s(\n%s)\n\n' % (attr_name,
                                              attr_type,
                                              format_props(attr_props, attr_scope, ignore)))
        subject_relations = [(rel[0].type, # rtype
                              [r.type for r in rel[1]], # etype targets
                              properties_from(entity.rdef(rel[0].type), permissionshandler)) # properties
                             for rel in sorted(entity.relation_definitions())
                             if rel[2] == 'subject']
        for rtype, targettypes, props in subject_relations:
            targettypes = [etype for etype in targettypes
                           if etype not in ignore]
            if not targettypes:
                continue
            if len(targettypes) == 1:
                ttypes_out = quoted(targettypes[0])
            else:
                multirtypes.append(rtype)
                continue
            output.append('\t%s = SubjectRelation(%s\n%s)\n\n' %
                       (rtype, ttypes_out, format_props(props, attr_scope, ignore)) )
        output.append('\n\n')

    # Handle multi-rdef relation types
    for rtype in sorted(multirtypes):
        for (subjectschema, objectschema), relation in sorted(schema[rtype].rdefs.iteritems()):
            subjetype = subjectschema.type
            objetype = objectschema.type
            output.append('class %s_%s_%s(RelationDefinition):\n' %
                          (rtype, subjetype.lower(), objetype.lower()))
            output.append('\tname = %s\n' % quoted(rtype))
            output.append('\tsubject = %s\n' % quoted(subjetype))
            output.append('\tobject = %s\n' % quoted(objetype))
            output.append(format_props(properties_from(relation, permissionshandler),
                                       class_scope, ignore))
            output.append('\n\n')
    return ''.join(output)

def schema2file(schema, output, permissionshandler, ignore=()):
    """Save schema description of schema find
    in directory schema_dir into output file"""
    with open(output, 'w') as description_file:
        description_file.write(schema2descr(schema, permissionshandler, ignore))

def schema_diff(schema1, schema2, permissionshandler=nullhandler, diff_tool=None, ignore=()):
    """schema 1 and 2 are CubicwebSchema
    diff_tool is the name of tool use for file comparison
    """
    tmpdir = tempfile.mkdtemp()
    output1 = os.path.join(tmpdir, "schema1.txt")
    output2 = os.path.join(tmpdir, "schema2.txt")
    schema2file(schema1, output1, permissionshandler, ignore)
    schema2file(schema2, output2, permissionshandler, ignore)
    if diff_tool:
        cmd = "%s %s %s" %(diff_tool,
                           output1, output2)
        process = subprocess.Popen(cmd, shell=True)
    else:
        print("description files save in %s and %s" % (output1, output2))
    print(output1, output2)
    return output1, output2

if __name__ == "__main__":
    from optparse import OptionParser

    parser = OptionParser()
    parser.add_option("-f", "--first-schema", dest="schema1",
                      help= "Specify the directory of the first schema")
    parser.add_option("-s", "--second-schema", dest="schema2",
                      help= "Specify the directory of the second schema")
    parser.add_option("-d", "--diff-tool", dest="diff_tool",
                      help= "Specify the name of the diff tool")
    (options, args) = parser.parse_args()
    if options.schema1 and options.schema2:
        schema1 = SchemaLoader().load([options.schema1])
        schema2 = SchemaLoader().load([options.schema2])
        output1, output2 = schema_diff(schema1, schema2)
        if options.diff_tool:
            cmd = "%s %s %s" %(options.diff_tool,
                               output1, output2)
            process = subprocess.Popen(cmd, shell=True)
        else:
            print("description files save in %s and %s" % (output1, output2))
    else:
        parser.error("An input file name must be specified")



#!/usr/bin/env python
# pylint: disable-msg=W0404,W0622,W0704,W0613,W0152
# copyright 2004-2010 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of yams.
#
# yams is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# yams is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with yams. If not, see <http://www.gnu.org/licenses/>.
"""Generic Setup script, takes package info from __pkginfo__.py file.

"""
__docformat__ = "restructuredtext en"

import os
import os.path as osp
import sys
import shutil
from io import open

from setuptools import setup, find_packages


here = osp.abspath(osp.dirname(__file__))

pkginfo = {}
with open(osp.join(here, '__pkginfo__.py')) as f:
    exec(f.read(), pkginfo)

# Get the long description from the relevant file
with open(osp.join(here, 'README'), encoding='utf-8') as f:
    long_description = f.read()

kwargs = {}
if 'subpackage_of' in pkginfo:
    kwargs['namespace_packages'] = [pkginfo['subpackage_of']]

setup(
    name=pkginfo.get('distname', pkginfo['modname']),
    version=pkginfo['version'],
    license=pkginfo['license'],
    description=pkginfo['description'],
    long_description=long_description,
    author=pkginfo['author'],
    author_email=pkginfo['author_email'],
    url=pkginfo['web'],
    classifiers=pkginfo.get('classifiers', []),
    packages=find_packages(exclude=['test*']),
    install_requires=pkginfo.get('install_requires'),
    scripts=pkginfo.get('scripts', []),
    ext_modules=pkginfo.get('ext_modules'),
    **kwargs
)



# copyright 2004-2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of yams.
#
# yams is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# yams is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with yams. If not, see <http://www.gnu.org/licenses/>.
"""Some common constraint classes."""

__docformat__ = "restructuredtext en"

import re
import decimal
import operator
import json
import datetime
import warnings

from six import string_types, text_type, binary_type

from logilab.common.deprecation import class_renamed

import yams
from yams import BadSchemaDefinition
from yams.interfaces import IConstraint, IVocabularyConstraint

_ = text_type


class ConstraintJSONEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, Attribute):
            return {'__attribute__': obj.attr}
        if isinstance(obj, NOW):
            d = obj.offset
            if d is not None:
                d = {'days': d.days, 'seconds': d.seconds, 'microseconds': d.microseconds}
            return {'__now__': True, 'offset': d}
        if isinstance(obj, TODAY):
            d = obj.offset
            if d is not None:
                d = {'days': d.days, 'seconds': d.seconds, 'microseconds': d.microseconds}
            return {'__today__': True, 'offset': d, 'type': obj.type}
        return super(ConstraintJSONEncoder, self).default(obj)


def _json_object_hook(dct):
    if '__attribute__' in dct:
        return Attribute(dct['__attribute__'])
    if '__now__' in dct:
        if dct['offset'] is not None:
            offset = datetime.timedelta(**dct['offset'])
        else:
            offset = None
        return NOW(offset)
    if '__today__' in dct:
        if dct['offset'] is not None:
            offset = datetime.timedelta(**dct['offset'])
        else:
            offset = None
        return TODAY(offset=offset, type=dct['type'])
    return dct


def cstr_json_dumps(obj):
    return text_type(ConstraintJSONEncoder(sort_keys=True).encode(obj))

cstr_json_loads = json.JSONDecoder(object_hook=_json_object_hook).decode


def _message_value(boundary):
    if isinstance(boundary, Attribute):
        return boundary.attr
    return boundary


class BaseConstraint(object):
    """base class for constraints"""
    __implements__ = IConstraint

    def __init__(self, msg=None):
        self.msg = msg

    def check_consistency(self, subjschema, objschema, rdef):
        pass

    def type(self):
        return self.__class__.__name__

    def serialize(self):
        """called to make persistent valuable data of a constraint"""
        return cstr_json_dumps({u'msg': self.msg})

    @classmethod
    def deserialize(cls, value):
        """called to restore serialized data of a constraint. Should return
        a `cls` instance
        """
        value = value.strip()
        if value and value != 'None':
            d = cstr_json_loads(value)
        else:
            d = {}
        return cls(**d)

    def failed_message(self, key, value, entity=None):
        if entity is None:
            warnings.warn('[yams 0.44] failed message should now be given entity has argument.',
                          DeprecationWarning, stacklevel=2)
        if self.msg:
            return self.msg, {}
        return self._failed_message(entity, key, value)

    def _failed_message(self, entity, key, value):
        return _('%(KEY-cstr)s constraint failed for value %(KEY-value)r'), {
            key + '-cstr': self,
            key + '-value': value}

    def __eq__(self, other):
        return (self.type(), self.serialize()) == (other.type(), other.serialize())

    def __ne__(self, other):
        return not self == other

    def __hash__(self):
        return hash((self.type(), self.serialize()))

    def __lt__(self, other):
        return NotImplemented


    def __eq__(self, other):
        return (self.type(), self.serialize()) == (other.type(), other.serialize())

    def __ne__(self, other):
        return not self == other

    def __hash__(self):
        return hash((self.type(), self.serialize()))


# possible constraints ########################################################

class UniqueConstraint(BaseConstraint):
    """object of relation must be unique"""

    def __str__(self):
        return 'unique'

    def check_consistency(self, subjschema, objschema, rdef):
        if not objschema.final:
            raise BadSchemaDefinition("unique constraint doesn't apply to non "
                                      "final entity type")

    def check(self, entity, rtype, values):
        """return true if the value satisfy the constraint, else false"""
        return True


class SizeConstraint(BaseConstraint):
    """the string size constraint :

    if max is not None the string length must not be greater than max
    if min is not None the string length must not be shorter than min
    """

    def __init__(self, max=None, min=None, msg=None):
        super(SizeConstraint, self).__init__(msg)
        assert (max is not None or min is not None), "No max or min"
        if min is not None:
            assert isinstance(min, int), 'min must be an int, not %r' % min
        if max is not None:
            assert isinstance(max, int), 'max must be an int, not %r' % max
        self.max = max
        self.min = min

    def __str__(self):
        res = 'size'
        if self.max is not None:
            res = '%s <= %s' % (res, self.max)
        if self.min is not None:
            res = '%s <= %s' % (self.min, res)
        return res

    def check_consistency(self, subjschema, objschema, rdef):
        if not objschema.final:
            raise BadSchemaDefinition("size constraint doesn't apply to non "
                                      "final entity type")
        if objschema not in ('String', 'Bytes', 'Password'):
            raise BadSchemaDefinition("size constraint doesn't apply to %s "
                                      "entity type" % objschema)
        if self.max:
            for cstr in rdef.constraints:
                if cstr.__class__ is StaticVocabularyConstraint:
                    for value in cstr.values:
                        if len(value) > self.max:
                            raise BadSchemaDefinition(
                                'size constraint set to %s but vocabulary '
                                'contains string of greater size' % self.max)

    def check(self, entity, rtype, value):
        """return true if the value is in the interval specified by
        self.min and self.max
        """
        if self.max is not None:
            if len(value) > self.max:
                return False
        if self.min is not None:
            if len(value) < self.min:
                return False
        return True

    def _failed_message(self, entity, key, value):
        if self.max is not None and len(value) > self.max:
            return _('value should have maximum size of %(KEY-max)s but found %(KEY-size)s'), {
                key + '-max': self.max,
                key + '-size': len(value)}
        if self.min is not None and len(value) < self.min:
            return _('value should have minimum size of %(KEY-min)s but found %(KEY-size)s'), {
                key + '-min': self.min,
                key + '-size': len(value)}
        assert False, 'shouldnt be there'

    def serialize(self):
        """simple text serialization"""
        return cstr_json_dumps({u'min': self.min, u'max': self.max,
                                u'msg': self.msg})

    @classmethod
    def deserialize(cls, value):
        """simple text deserialization"""
        try:
            d = cstr_json_loads(value)
            return cls(**d)
        except ValueError:
            kwargs = {}
            for adef in value.split(','):
                key, val = [w.strip() for w in adef.split('=')]
                assert key in ('min', 'max')
                kwargs[str(key)] = int(val)
            return cls(**kwargs)


class RegexpConstraint(BaseConstraint):
    """specifies a set of allowed patterns for a string value"""
    __implements__ = IConstraint

    def __init__(self, regexp, flags=0, msg=None):
        """
        Construct a new RegexpConstraint.

        :Parameters:
         - `regexp`: (str) regular expression that strings must match
         - `flags`: (int) flags that are passed to re.compile()
        """
        super(RegexpConstraint, self).__init__(msg)
        self.regexp = regexp
        self.flags = flags
        self._rgx = re.compile(regexp, flags)

    def __str__(self):
        return 'regexp %s' % self.regexp

    def check_consistency(self, subjschema, objschema, rdef):
        if not objschema.final:
            raise BadSchemaDefinition("regexp constraint doesn't apply to non "
                                      "final entity type")
        if objschema not in ('String', 'Password'):
            raise BadSchemaDefinition("regexp constraint doesn't apply to %s "
                                      "entity type" % objschema)

    def check(self, entity, rtype, value):
        """return true if the value maches the regular expression"""
        return self._rgx.match(value, self.flags)

    def _failed_message(self, entity, key, value):
        return _("%(KEY-value)r doesn't match the %(KEY-regexp)r regular expression"), {
            key + '-value': value,
            key + '-regexp': self.regexp}

    def serialize(self):
        """simple text serialization"""
        return cstr_json_dumps({u'regexp': self.regexp, u'flags': self.flags,
                                u'msg': self.msg})

    @classmethod
    def deserialize(cls, value):
        """simple text deserialization"""
        try:
            d = cstr_json_loads(value)
            return cls(**d)
        except ValueError:
            regexp, flags = value.rsplit(',', 1)
            return cls(regexp, int(flags))

    def __deepcopy__(self, memo):
        return RegexpConstraint(self.regexp, self.flags)


OPERATORS = {
    '<=': operator.le,
    '<': operator.lt,
    '>': operator.gt,
    '>=': operator.ge,
}


class BoundaryConstraint(BaseConstraint):
    """the int/float bound constraint :

    set a minimal or maximal value to a numerical value
    """
    __implements__ = IConstraint

    def __init__(self, op, boundary=None, msg=None):
        super(BoundaryConstraint, self).__init__(msg)
        assert op in OPERATORS, op
        self.operator = op
        self.boundary = boundary

    def __str__(self):
        return 'value %s' % self.serialize()

    def check_consistency(self, subjschema, objschema, rdef):
        if not objschema.final:
            raise BadSchemaDefinition("bound constraint doesn't apply to non "
                                      "final entity type")

    def check(self, entity, rtype, value):
        """return true if the value satisfies the constraint, else false"""
        boundary = actual_value(self.boundary, entity)
        if boundary is None:
            return True
        return OPERATORS[self.operator](value, boundary)

    def _failed_message(self, entity, key, value):
        return "value %%(KEY-value)s must be %s %%(KEY-boundary)s" % self.operator, {
            key + '-value': value,
            key + '-boundary': _message_value(actual_value(self.boundary, entity))}

    def serialize(self):
        """simple text serialization"""
        return cstr_json_dumps({u'op': self.operator, u'boundary': self.boundary,
                                u'msg': self.msg})

    @classmethod
    def deserialize(cls, value):
        """simple text deserialization"""
        try:
            d = cstr_json_loads(value)
            return cls(**d)
        except ValueError:
            op, boundary = value.split(' ', 1)
            return cls(op, eval(boundary))


BoundConstraint = class_renamed('BoundConstraint', BoundaryConstraint)
BoundConstraint.type = lambda x: 'BoundaryConstraint'

_("value %(KEY-value)s must be < %(KEY-boundary)s")
_("value %(KEY-value)s must be > %(KEY-boundary)s")
_("value %(KEY-value)s must be <= %(KEY-boundary)s")
_("value %(KEY-value)s must be >= %(KEY-boundary)s")


class IntervalBoundConstraint(BaseConstraint):
    """an int/float bound constraint :

    sets a minimal and / or a maximal value to a numerical value
    This class replaces the BoundConstraint class
    """
    __implements__ = IConstraint

    def __init__(self, minvalue=None, maxvalue=None, msg=None):
        """
        :param minvalue: the minimal value that can be used
        :param maxvalue: the maxvalue value that can be used
        """
        assert not (minvalue is None and maxvalue is None)
        super(IntervalBoundConstraint, self).__init__(msg)
        self.minvalue = minvalue
        self.maxvalue = maxvalue

    def __str__(self):
        return 'value [%s]' % self.serialize()

    def check_consistency(self, subjschema, objschema, rdef):
        if not objschema.final:
            raise BadSchemaDefinition("interval bound constraint doesn't apply "
                                      "to non final entity type")

    def check(self, entity, rtype, value):
        minvalue = actual_value(self.minvalue, entity)
        if minvalue is not None and value < minvalue:
            return False
        maxvalue = actual_value(self.maxvalue, entity)
        if maxvalue is not None and value > maxvalue:
            return False
        return True

    def _failed_message(self, entity, key, value):
        if self.minvalue is not None and value < actual_value(self.minvalue, entity):
            return _("value %(KEY-value)s must be >= %(KEY-boundary)s"), {
                key + '-value': value,
                key + '-boundary': _message_value(self.minvalue)}
        if self.maxvalue is not None and value > actual_value(self.maxvalue, entity):
            return _("value %(KEY-value)s must be <= %(KEY-boundary)s"), {
                key + '-value': value,
                key + '-boundary': _message_value(self.maxvalue)}
        assert False, 'shouldnt be there'

    def serialize(self):
        """simple text serialization"""
        return cstr_json_dumps({u'minvalue': self.minvalue, u'maxvalue': self.maxvalue,
                                u'msg': self.msg})

    @classmethod
    def deserialize(cls, value):
        """simple text deserialization"""
        try:
            d = cstr_json_loads(value)
            return cls(**d)
        except ValueError:
            minvalue, maxvalue = value.split(';')
            return cls(eval(minvalue), eval(maxvalue))


class StaticVocabularyConstraint(BaseConstraint):
    """Enforces a predefined vocabulary set for the value."""
    __implements__ = IVocabularyConstraint

    def __init__(self, values, msg=None):
        super(StaticVocabularyConstraint, self).__init__(msg)
        self.values = tuple(values)

    def __str__(self):
        return 'value in (%s)' % u', '.join(repr(text_type(word)) for word in self.vocabulary())

    def check(self, entity, rtype, value):
        """return true if the value is in the specific vocabulary"""
        return value in self.vocabulary(entity=entity)

    def _failed_message(self, entity, key, value):
        if isinstance(value, string_types):
            value = u'"%s"' % text_type(value)
            choices = u', '.join('"%s"' % val for val in self.values)
        else:
            choices = u', '.join(text_type(val) for val in self.values)
        return _('invalid value %(KEY-value)s, it must be one of %(KEY-choices)s'), {
            key + '-value': value,
            key + '-choices': choices}

    def vocabulary(self, **kwargs):
        """return a list of possible values for the attribute"""
        return self.values

    def serialize(self):
        """serialize possible values as a json object"""
        return cstr_json_dumps({u'values': self.values, u'msg': self.msg})

    @classmethod
    def deserialize(cls, value):
        """deserialize possible values from a csv list of evaluable strings"""
        try:
            values = cstr_json_loads(value)
            return cls(**values)
        except ValueError:
            values = [eval(w) for w in re.split('(?<!,), ', value)]
            if values and isinstance(values[0], string_types):
                values = [v.replace(',,', ',') for v in values]
            return cls(values)


class FormatConstraint(StaticVocabularyConstraint):

    regular_formats = (_('text/rest'),
                       _('text/markdown'),
                       _('text/html'),
                       _('text/plain'),
                       )

    def __init__(self, msg=None, **kwargs):
        values = self.regular_formats
        super(FormatConstraint, self).__init__(values, msg=msg)

    def check_consistency(self, subjschema, objschema, rdef):
        if not objschema.final:
            raise BadSchemaDefinition("format constraint doesn't apply to non "
                                      "final entity type")
        if not objschema == 'String':
            raise BadSchemaDefinition("format constraint only apply to String")


FORMAT_CONSTRAINT = FormatConstraint()


class MultipleStaticVocabularyConstraint(StaticVocabularyConstraint):
    """Enforce a list of values to be in a predefined set vocabulary."""
    # XXX never used
    def check(self, entity, rtype, values):
        """return true if the values satisfy the constraint, else false"""
        vocab = self.vocabulary(entity=entity)
        for value in values:
            if value not in vocab:
                return False
        return True


# special classes to be used w/ constraints accepting values as argument(s):
# IntervalBoundConstraint

def actual_value(value, entity):
    if hasattr(value, 'value'):
        return value.value(entity)
    return value


class Attribute(object):
    def __init__(self, attr):
        self.attr = attr

    def __str__(self):
        return '%s(%r)' % (self.__class__.__name__, self.attr)

    def value(self, entity):
        return getattr(entity, self.attr)


class NOW(object):
    def __init__(self, offset=None):
        self.offset = offset

    def __str__(self):
        return '%s(%r)' % (self.__class__.__name__, self.offset)

    def value(self, entity):
        now = yams.KEYWORD_MAP['Datetime']['NOW']()
        if self.offset:
            now += self.offset
        return now


class TODAY(object):
    def __init__(self, offset=None, type='Date'):
        self.offset = offset
        self.type = type

    def __str__(self):
        return '%s(%r, %r)' % (self.__class__.__name__, self.offset, self.type)

    def value(self, entity):
        now = yams.KEYWORD_MAP[self.type]['TODAY']()
        if self.offset:
            now += self.offset
        return now


# base types checking functions ################################################

def check_string(eschema, value):
    """check value is an unicode string"""
    return isinstance(value, text_type)


def check_password(eschema, value):
    """check value is an encoded string"""
    return isinstance(value, binary_type)


def check_int(eschema, value):
    """check value is an integer"""
    try:
        int(value)
    except ValueError:
        return False
    return True


def check_float(eschema, value):
    """check value is a float"""
    try:
        float(value)
    except ValueError:
        return False
    return True


def check_decimal(eschema, value):
    """check value is a Decimal"""
    try:
        decimal.Decimal(value)
    except (TypeError, decimal.InvalidOperation):
        return False
    return True


def check_boolean(eschema, value):
    """check value is a boolean"""
    return isinstance(value, int)


def check_file(eschema, value):
    """check value has a getvalue() method (e.g. StringIO or cStringIO)"""
    return hasattr(value, 'getvalue')


def yes(*args, **kwargs):
    """dunno how to check"""
    return True


BASE_CHECKERS = {
    'Date': yes,
    'Time': yes,
    'Datetime': yes,
    'TZTime': yes,
    'TZDatetime': yes,
    'Interval': yes,
    'String': check_string,
    'Int': check_int,
    'BigInt': check_int,
    'Float': check_float,
    'Decimal': check_decimal,
    'Boolean': check_boolean,
    'Password': check_password,
    'Bytes': check_file,
}

BASE_CONVERTERS = {
    'String': text_type,
    'Password': binary_type,
    'Int': int,
    'BigInt': int,
    'Float': float,
    'Boolean': bool,
    'Decimal': decimal.Decimal,
}


def patch_sqlite_decimal():
    """patch Decimal checker and converter to bypass SQLITE Bug
    (SUM of Decimal return float in SQLITE)"""

    def convert_decimal(value):
        # XXX issue a warning
        if isinstance(value, float):
            value = str(value)
        return decimal.Decimal(value)

    def check_decimal(eschema, value):
        """check value is a Decimal"""
        try:
            if isinstance(value, float):
                return True
            decimal.Decimal(value)
        except (TypeError, decimal.InvalidOperation):
            return False
        return True

    global BASE_CONVERTERS
    BASE_CONVERTERS['Decimal'] = convert_decimal
    global BASE_CHECKERS
    BASE_CHECKERS["Decimal"] = check_decimal
